#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Helps migrate picture settings form F-Spot to DigiKam
# 
# Original idea and script from
# http://blog.mekk.waw.pl/archives/12-Moving-images-from-F-Spot-to-digiKam.html
# Adapted and extended by Roland Geider <roland@geider.net>, 2011
#
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


def needed_module(name, package, error):
    print("Error! This script uses %s. Please install the following" % name)
    print('package and try again: "%s"' % package)
    print("(name of package in a debian based system)")
    print("The error message from python is: '%s'" % error)
    print("")
    sys.exit()

import sys
import os
import shutil
import time
import logging
import optparse
import urlparse

# If we are using python 3, change the import
if sys.version_info[0] == 3:
    import urllib.parse as urllib
else:
    import urllib

try:
     import dbus
except ImportError as e:
    needed_module('DBUS', 'python-dbus', e)
    
try: 
    from sqlalchemy import create_engine, MetaData
    from sqlalchemy import Column, Integer, Unicode, ForeignKey
    from sqlalchemy.orm import sessionmaker, relationship
    from sqlalchemy.orm.exc import MultipleResultsFound, NoResultFound
    from sqlalchemy.ext.declarative import declarative_base
    from sqlalchemy.interfaces import PoolListener
except ImportError as e:
    needed_module('SQLalchemy', 'python-sqlalchemy', e)




parser = optparse.OptionParser(usage="""python %prog DIGIKAM_DB [options]

This script helps you migrate from F-Spot to Digikam. Before doing
anything it makes a backup of the SQLite files so if anything goes wrong
you can revert all changes.

You must specify DIGIKAM_DB, the folder where Digikam's database is located,
relative to your home folder (Pictures for /home/your_username/Pictures)
This is configured under
  Settings > Configure digikam > Database [SQLite] > path


This was tested with the table structures for

  * F-Spot 0.7.3 and 0.8.2
  * Digikam 1.2.0, 1.9.0 and 2.1.1

For more info please consult the README file
""")

parser.add_option('--fspot-folder',
                    dest="custom_fspot_db",
                    help="If the f-spot database is not in the default location you can change the "\
                        "folder where it is. Relative to your home folder. It is assumed that the "\
                        "file is called photos.db.")
parser.add_option('--ignore-ratings',
                    action="store_true",
                    dest="ignore_ratings",
                    default=False,
                    help="Set this option to skip F-Spot's ratings on pictures. This will speed up "\
                        "the process a slight bit")
parser.add_option('--ignore-comments',
                    action="store_true",
                    dest="ignore_comments",
                    default=False,
                    help="Set this option to skip F-Spot's comments on pictures. As with "\
                        "--ignore-ratings, this will speed up the process, and if the comments were "\
                        "set by your camera, digikam will be able to read them anyway.")
parser.add_option('--ignore-versioning',
                    action="store_true",
                    dest="ignore_versioning",
                    default=False,
                    help="Set this option to skip F-Spot's grouping of pictures (versioning). "\
                        "This is only supported by Digikam starting with 2.0. If you use an "\
                        "older version, migration of picture versioning is disabled automatically.")
parser.add_option('--reparent-tags',
                    action="store_true",
                    dest="reparent_tags",
                    default=False,
                    help="Reparent tags already existing in Digikam to get the same tag tree "
                        "as in F-Spot. This is useful if the tags have already been extracted "
                        "as a flat list when importing the photos.")
parser.add_option('-v', '--verbose',
                    action="store_true",
                    dest="debugging",
                    default=False,
                    help="will output debugging information. Warning: this will slow down the "\
                        "conversion.")
parser.add_option('--verbose-db',
                    action="store_true",
                    dest="debugging_db",
                    default=False,
                    help="output SQLalchemy's database debug information. Warning: *extremely* "\
                        "chatty!!")
parser.add_option('--dry-run',
                    action="store_true",
                    dest="dry_run",
                    default=False,
                    help="will not change anything on the digikam database (does a rollback "\
                        "at the end). A backup of the Database will still be made.")
                        
options, remainder = parser.parse_args()


if len(remainder) == 0:
    parser.print_help()
    sys.exit()
elif len(remainder) != 1:
    print("""You must specify DIGIKAM_DB, the folder where Digikam's database is located,
relative to your home folder (e.g. "Pictures" for "/home/your_username/Pictures"). 
This is configured under Settings > Configure digikam > Database [SQLite] > 
path. It is assumed that the file is called digikam4.db.""")
    sys.exit()
 
# Read the settings from the options object
custom_fspot_db       = options.custom_fspot_db
import_fspot_rating   = not options.ignore_ratings
import_fspot_comments = not options.ignore_comments
import_fspot_picture_versions = not options.ignore_versioning 
reparent_tags         = options.reparent_tags
if options.debugging:
    logging_level     = logging.DEBUG
else:
    logging_level     = logging.WARNING
digikam_folder        = remainder[0]
sqlalchemy_debug      = options.debugging_db
dry_run               = options.dry_run


# Initialise the statistics
stats_number_of_tags = 0
stats_number_of_picures = 0
stats_number_of_picures_error = 0


logging.basicConfig(level=logging_level,
                    format = '%(levelname)s: %(message)s')

# DB locations
home = os.environ['HOME']
fspot_db = "%(home)s/.config/f-spot/photos.db" % locals()
if custom_fspot_db:
    fspot_db = "%s/%s/photos.db" % (locals()['home'], custom_fspot_db)
digikam_db = "%s/%s/digikam4.db" % (locals()['home'], digikam_folder)


# Do some sanity checks before we start meddling with the DBs
def check_file(what, path):
    print("Checking for %s database file" % what)
    if not os.path.exists(path):
        
        print("-> ERROR! Could not find the %s database at %s" % (what, path))
        print("   Are you sure this is the correct location?")
        print("")
        print("Exiting")
        sys.exit()
    else:
        print("-> OK")

check_file('F-Spot', fspot_db)
check_file('Digikam', digikam_db)


# Backup the digikam DB, in case we mess up
backup_db = digikam_db + '-' + time.strftime("%Y-%m-%d_%H%M") + '.bak'
print("Making a backup of the Digikam database...")

try:
    shutil.copy(digikam_db, backup_db)
    pass
except IOError as e:
    print("-> ERROR, could not make a backup. Details follow: %s" % e)

print("-> OK, the backup file is located at: %s" % backup_db)
print("   If there are problems after running this script rename to %s" % digikam_db)


# Setup a DBUS helper function
#
# Yes! we are murking with Dbus and devices' UUIDs because we need to map digikam's
# volume identifier to a mountpoint in the system, as fspot uses
bus = dbus.SystemBus()
ud_manager_obj = bus.get_object("org.freedesktop.UDisks", "/org/freedesktop/UDisks")
ud_manager = dbus.Interface(ud_manager_obj, 'org.freedesktop.UDisks')

uuid_mapper = {}

for dev in ud_manager.EnumerateDevices():
    device_obj = bus.get_object("org.freedesktop.UDisks", dev)
    device_props = dbus.Interface(device_obj, dbus.PROPERTIES_IFACE)
    
    # make a mapping mountpoint ==> UUID
    if device_props.Get('org.freedesktop.UDisks.Device', "IdUuid") and\
       device_props.Get('org.freedesktop.UDisks.Device', "DeviceMountPaths"):
        uuid_mapper[str(device_props.Get('org.freedesktop.UDisks.Device', "DeviceMountPaths")[0])] = \
            str(device_props.Get('org.freedesktop.UDisks.Device', "IdUuid"))


#
# Set up the F-Spot database
#
class SetTextFactory(PoolListener):
    def connect(self, dbapi_con, con_record):
        dbapi_con.text_factory = str

fspot_engine = create_engine('sqlite:////' + fspot_db, echo=sqlalchemy_debug, listeners=[SetTextFactory()])
fspot_metadata = MetaData()
Base = declarative_base(metadata=fspot_metadata)



class FSpotPhotoTags(Base):
    __tablename__ = 'photo_tags'
    
    # Columns
    photo_id = Column(Integer, ForeignKey('photos.id'), primary_key = True)
    tag_id = Column(Integer, ForeignKey('tags.id'), primary_key = True)
    

class FSpotTag(Base):
    __tablename__ = 'tags'
    autoload_with = fspot_engine
    
    # Columns
    id = Column(Integer, primary_key=True)
    name = Column(Unicode)
    category_id = Column(Integer, ForeignKey('tags.id'))
    sort_priority = Column(Integer)
    icon = Column(Unicode)
    
    # Relationships
    tags = relationship(FSpotPhotoTags, backref = "tag")


class FSpotPhoto(Base):
    __tablename__ = 'photos'
    
    # Columns
    id = Column(Integer, primary_key=True)
    time = Column(Integer)
    base_uri = Column(Unicode)
    filename = Column(Unicode)
    description = Column(Unicode)
    roll_id = Column(Integer)
    default_version_id = Column(Integer)#, ForeignKey('photo_versions.version_id'))
    rating = Column(Integer)
    
    # Relationships
    tags = relationship(FSpotPhotoTags, backref = "foto")


class FSpotPhotoVersion(Base):
    __tablename__ = 'photo_versions'
    
    # Columns
    photo_id = Column(Integer, ForeignKey('photos.id'), primary_key=True)
    version_id = Column(Integer, primary_key=True)
    base_uri = Column(Unicode)
    filename = Column(Unicode)
    
    # Relationships
    photo = relationship(FSpotPhoto, backref = "versions")


# Start a session
FSpotSession = sessionmaker(autoflush = True)
FSpotSession.configure(bind = fspot_engine)
fspot_session = FSpotSession()


#
# Digikam database
#

digikam_engine = create_engine('sqlite:////' + digikam_db, echo=sqlalchemy_debug, listeners=[SetTextFactory()])
digikam_metadata = MetaData()
Base = declarative_base(metadata=digikam_metadata)


class DigikamAlbumRoot(Base):
    __tablename__ = 'AlbumRoots'
    
    # Columns
    id = Column(Integer, primary_key = True)
    identifier = Column(Unicode)
    specificPath = Column(Unicode)
    

class DigikamAlbum(Base):
    __tablename__ = 'Albums'
    
    # Columns
    id = Column (Integer, ForeignKey("Images.album"), primary_key = True)
    relativePath = Column(Unicode)
    albumRoot = Column(Unicode, ForeignKey("AlbumRoots.id"))
    date = Column(Unicode)
    caption = Column(Unicode)
    
    # Relationships
    album_root = relationship(DigikamAlbumRoot, backref = "albums")



class DigikamPhoto(Base):
    __tablename__ = "Images"
    
    # Tables
    id = Column (Integer, primary_key = True)
    name = Column(Unicode)
    album = Column(Integer, ForeignKey("Albums.id"))
    
    # Relationships
    album_mapping = relationship(DigikamAlbum,
                        backref = "photo",
                        primaryjoin = album == DigikamAlbum.id)


class DigikamTag(Base):
    __tablename__ = "Tags"
    
    def __init__(self, name, parent_id=0, id = None):
        self.name = name
        self.icon = 0
        self.pid = parent_id
        if id:
            self.id = id
    
    # Columns
    id = Column (Integer, primary_key = True)
    pid = Column (Integer, ForeignKey('Tags.id'))
    name = Column(Unicode)
    icon = Column(Integer)


class DigikamImageTags(Base):
    __tablename__ = "ImageTags"
    
    def __init__(self, image_id = None, tag_id = None):
        if image_id:
            self.imageid = image_id
        if tag_id:
            self.tagid = tag_id
    
    # Columns
    imageid = Column(Integer, ForeignKey('Images.id'), primary_key = True)
    tagid = Column(Integer, ForeignKey('Tags.id'), primary_key = True)
    
    # Relationships
    photo = relationship(DigikamPhoto, backref = "tag")


class ImageInformation(Base):
    __tablename__ = "ImageInformation"
    
    def __init__(self, image_id = None, rating = None):
        if image_id:
            self.imageid = image_id
        if rating:
            self.rating = rating
            
    # Columns
    imageid = Column(Integer, ForeignKey('Images.id'), primary_key = True)
    rating = Column(Integer)
    
    # Relationships
    photo = relationship(DigikamPhoto, backref = "information")

class ImageComment(Base):
    __tablename__ = "ImageComments"
    
    def __init__(self, image_id = None, comment = None):
        if image_id:
            self.imageid = image_id
        if comment:
            self.comment = comment
        self.language = 'x-default'
        self.type = 1
            
    # Columns
    id = Column(Integer, primary_key = True)
    imageid = Column(Integer,ForeignKey('Images.id'), primary_key = True)
    type = Column(Integer)
    language = Column(Unicode)
    author = Column(Unicode)
    comment = Column(Unicode)
    imageid = Column(Integer, ForeignKey('Images.id'), primary_key = True)
    
    # Relationships
    photo = relationship(DigikamPhoto, backref = "comments")


class ImageRelations(Base):
    __tablename__ = "ImageRelations"
    
    def __init__(self, subject = None, object = None, type = None):
        if subject:
            self.subject = subject
        if object:
            self.object = object
        if type:
            self.type = type
        else:
            self.type = 2
        
            
    # Columns
    subject = Column(Integer, ForeignKey('Images.id'), primary_key = True)
    object = Column(Integer, ForeignKey('Images.id'), primary_key = True)
    type = Column(Integer, primary_key = True)

class DigikamDbInformation(Base):
    __tablename__ = 'Settings'
    
    # Columns
    keyword = Column(Unicode, primary_key = True)
    value = Column(Unicode, primary_key = True)
    

# Start a session
DigikamSession = sessionmaker(autoflush = True)
DigikamSession.configure(bind = digikam_engine)
digikam_session = DigikamSession()


# See if we are working with an 1.x database or a new 2.x one
# This will decide if we try to migrate the versions
db_info = digikam_session.query(DigikamDbInformation).filter_by(keyword = 'DBVersion').one()
db_info.value = int(db_info.value)

digikam_2x = False
if db_info.value >= 6:
    digikam_2x = True

# Ask the user to continue if using a newer version of Digikam
if db_info.value > 6:
    print ("""
You are using a version of Digikam that has a newer database structure than
what this script was tested with. While we only do "basic" things, it could
be that the results are somewhat off or the script chokes on some of the
differences. This should not sound scary, there is a backup of the digikam
database anyway! Continue?""")
    while True:
        try_with_new_digikam = raw_input("Please answer with 'Y' or 'N': ")
        if try_with_new_digikam.lower() == 'n':
            print ("OK, exiting")
            sys.exit()
        
        if try_with_new_digikam.lower() == 'y':
            print ("OK, will continue")
            break     

#
# OK, we have everything we need, let's start the migration!
#
print("")
print("Everything is set up, starting...")


# Cache the structure of the fspot tags
print("Copying tag structure from F-spot")
print("---------------------------------")
fspot_tag_cache = {}
digikam_tag_cache = {}
for fspot_tag in fspot_session.query(FSpotTag).all():
    digikam_tag = digikam_session.query(DigikamTag).filter_by(name = fspot_tag.name).first()
    logging.info("Creating missing Digikam tag %s" % fspot_tag.name)
    
    fspot_tag_parent = fspot_session.query(FSpotTag).filter_by(id = fspot_tag.category_id).first()
    fspot_tag_parent_id = 0
    if fspot_tag_parent:
        fspot_tag_parent_id = fspot_tag_parent.id
    
    fspot_tag_cache[fspot_tag.id] = {'name': fspot_tag.name,
                                    'digikam_tag': digikam_tag,
                                    'parent_id': fspot_tag_parent_id,
                                    'parent_name': fspot_tag_parent and fspot_tag_parent.name}
    if not digikam_tag:
        digikam_tag = DigikamTag(fspot_tag.name)
        digikam_session.add(digikam_tag)
    else:
        logging.info("F-Spot Tag %s already in Digikam, skiping..." % digikam_tag.name)
    
    digikam_tag_cache[fspot_tag.id] = digikam_tag


for fspot_tag in fspot_tag_cache:
    if fspot_tag_cache[fspot_tag]['parent_id']:
        digikam_tag_parent = digikam_session.query(DigikamTag).filter_by(name = fspot_tag_cache[fspot_tag]['parent_name']).first()
        digikam_tag = digikam_session.query(DigikamTag).filter_by(name = fspot_tag_cache[fspot_tag]['name']).first()
    
        if not fspot_tag_cache[fspot_tag]['digikam_tag']:
            digikam_session.merge(DigikamTag(fspot_tag_cache[fspot_tag]['name'],
                                            parent_id = digikam_tag_parent.id,
                                            id = digikam_tag.id))
        elif reparent_tags:
            logging.info("Setting parent for %s to %s.",
                         fspot_tag_cache[fspot_tag]['digikam_tag'].name,
                         digikam_tag_parent.name)
            fspot_tag_cache[fspot_tag]['digikam_tag'].pid = digikam_tag_parent.id

        stats_number_of_tags = stats_number_of_tags + 1
print("-> OK")
    

mount_point_cache = {}
def get_mount_point_from_uri(fspot_uri):
    """
    Returns the mount point given a URI from a picture form the f-spot
    database.
    
    For example: returns "/" or "/my/path/to" from "/my/path/to/my/pictures"
    """
    if mount_point_cache.get(fspot_uri, False):
        return mount_point_cache[fspot_uri]
        
    temp = ''
    longest = '/'
    for level in fspot_uri.strip('/').split('/'):
        temp = temp + "/" + level 
        if temp in uuid_mapper:
            longest = temp

    mount_point_cache[fspot_uri] = longest
        
    return longest


# Get all Digikam albumRoots
digikam_album_roots = {}
digikam_albums = digikam_session.query(DigikamAlbumRoot).all()
for root in digikam_albums:
    root_identifier_url = urlparse.urlsplit(root.identifier)
    if root_identifier_url.scheme == "networkshareid":
        path = urlparse.parse_qs(root_identifier_url.path)["?mountpath"][0]
        digikam_album_roots[path] = root.id
    else:
        digikam_album_roots[root.specificPath] = root.id



digikam_album_root_cache = {}
def get_digikam_album_from_uri(fspot_uri):
    """
    Returns the Digikam AlbumRoot ID and Album's relativePath given the URI
    of an F-Spot picture
    """
    if digikam_album_root_cache.get(fspot_uri, False):
        return digikam_album_root_cache[fspot_uri]
    
    # fspot_uri in the form "file:///home/user/Photos/2010/6/1/"
    base_uri = os.path.normpath(fspot_uri).replace('file:', '')
    mount_point = get_mount_point_from_uri(base_uri)
    
    # mount_point in the form "/home"
    if not mount_point:
        print("Could not find the mount point of %s." % (fspot_uri))
        print("Skipping...")
        print("")
        return False
    
    # Find the album starting with the path
    digikam_album_root = ''
    for album_root in digikam_album_roots:
        startswith = mount_point_cache[base_uri] + album_root
        startswith = startswith.replace('//', '/')

        if base_uri.startswith(startswith):
            digikam_album_root = album_root
    
    if not digikam_album_root:
        print("Could not locate the Digikam album corresponding to %s." % base_uri)
        print("Skipping...")
        print("")
        return False
    temp_root = mount_point_cache[base_uri] + digikam_album_root 
    album_relative_path = base_uri.partition(temp_root.replace('//', '/'))[2]

    temp_root = mount_point_cache[base_uri] + digikam_album_root 
    album_relative_path = base_uri.partition(temp_root.replace('//', '/'))[2]
    
    digikam_album_root_cache[fspot_uri] = {'album_root_id': digikam_album_roots[digikam_album_root],
                                            'album_relative_path': album_relative_path}
    return digikam_album_root_cache[fspot_uri]
    

def find_photo_in_digikam(fspot_photo):
    """
    Finds a picture in the digikam database given a f-spot one
    """
    
    # Find the Digikam album corresponding to this picture
    digikam_album_root = get_digikam_album_from_uri(fspot_photo.base_uri)
    
    if not digikam_album_root:
        return False
    
    
    
    # This should take care of pictures with a non-ASCII filename.
    # (Digikam doesn't quote the filename, while f-spot does)
    try:
        unquoted_filename = urllib.unquote(fspot_photo.filename)
        if sys.version_info[0] == 3:
            unquoted_filename = str(unquoted_filename, 'utf-8')
        else:
            unquoted_filename = unicode(unquoted_filename, 'utf-8')
    except UnicodeDecodeError:
        print("Could not decode the following filename: %s" % fspot_photo.filename)
        print("Skipping...")
        print("")
        
        return False
    
    
    # Try to find the same picture in Digikam
    try:
        digikam_photo = digikam_session.query(DigikamPhoto). \
                    join((DigikamAlbum, DigikamPhoto.album==DigikamAlbum.id)).\
                    filter(DigikamAlbum.albumRoot == digikam_album_root['album_root_id']). \
                    filter(DigikamAlbum.relativePath == digikam_album_root['album_relative_path']). \
                    filter(DigikamPhoto.name == unquoted_filename).one()
    except MultipleResultsFound:
        print("Found more than one picture. There was probably an error with the")
        print("path to the album. Details follow...")
        print("- Foto-ID:     %s" % fspot_photo.id)
        print("- Filename:    %s" % fspot_photo.filename)
        print("- Base folder: %s" % fspot_photo.base_uri)
        print("- Calculated digikam relative path: %s" % digikam_album_root['album_relative_path'])
        print("Skipping...")
        print("")
        
        return False
    
    except NoResultFound:    
        print("Found no corresponding picture in Digikam! Image orphaned? Details follow:")
        print("- Foto-ID:     %s" % fspot_photo.id)
        print("- Filename:    %s" % fspot_photo.filename)
        print("- Base folder: %s" % fspot_photo.base_uri)
        print("Skipping...")
        print("")
        
        return False
    
    return digikam_photo


def migrate_to_digikam(fspot_photo, digikam_picture=None):
    """
    Does the actual migrations, copying the tags, ratings, etc. from the f-spot
    database.
    """
    global stats_number_of_picures_error
    global stats_number_of_picures
    
    # Find the picture in the digikam database and return if not found
    if not digikam_picture:
        digikam_photo = find_photo_in_digikam(fspot_photo)
        
        if not digikam_photo:
            stats_number_of_picures_error = stats_number_of_picures_error + 1
            return False
        stats_number_of_picures = stats_number_of_picures + 1
    
    
    # Set the tags
    for fspot_tag in fspot_session.query(FSpotPhotoTags).filter_by(photo_id = fspot_photo.id):
        logging.info("Setting tag %s" % (fspot_tag.tag.name))
        try:
            digikam_tag = digikam_tag_cache[fspot_tag.tag.id]
        except KeyError:
            print("F-spot tag %s not in the digikam cache. Ignoring." % fspot_tag.tag.id)
            return False
        
        digikam_session.merge(DigikamImageTags(image_id = digikam_photo.id,
                                               tag_id = digikam_tag.id))

    # Set the rating
    if import_fspot_rating:
        if fspot_photo.rating:
            logging.info("Setting a rating of %d" % fspot_photo.rating)
            digikam_session.query(ImageInformation).filter_by(imageid = digikam_photo.id).update({'rating': fspot_photo.rating})
    
    # Set the comments
    if import_fspot_comments:
        if fspot_photo.description.strip():
            logging.info('Setting the following comment: "%s"' % (fspot_photo.description))
            digikam_session.merge(ImageComment(image_id = digikam_photo.id,
                                               comment = fspot_photo.description.strip()))
    return True

                
# Migrate the pictures and set the rating and comment of the parent
print("")
print("Processing the pictures. This might take a minute")
print("-------------------------------------------------")

for fspot_photo in fspot_session.query(FSpotPhoto).join(FSpotPhotoVersion).all():
    
    default_version_id = fspot_photo.default_version_id
    current_root = None
    migration_successful = True
    current_versions = []
    
    for version in fspot_photo.versions:
        # Set the filename of the current version to the parent picture and
        # import it as if it were a new picture, then, set it back to what it
        # was
        version_original_filename = fspot_photo.filename
        fspot_photo.filename = version.filename
        
        # Stop here if there was a problem with the migration
        if not migrate_to_digikam(fspot_photo):
            migration_successful = False
        fspot_photo.filename = version_original_filename
        
        # add the current version to our list
        current_versions.append(version)
    
        if version.version_id == default_version_id:
            current_root = version
    
    # If we are not working with a Digikam 2.x database or if the user didn't want to,
    # we finish here and don't copy the versions
    if not digikam_2x or not import_fspot_picture_versions or not migration_successful:
        continue
        
    # current_root is a FSpotPhotoVersion, which does not have an ID. We set it
    # here in case we need to output an error
    current_root.id = 'not available'
    
    digikam_root_photo = find_photo_in_digikam(current_root)
    if not digikam_root_photo:
        continue
    
    # now that we know the default version, we can iterate through the list
    # and set the relations between them
    for version in current_versions:
        
        digikam_version_photo = find_photo_in_digikam(version)
        if not digikam_version_photo:
            continue
        
        if digikam_root_photo.id != digikam_version_photo.id:
            digikam_session.merge(ImageRelations(subject = digikam_version_photo.id,
                                                object = digikam_root_photo.id))

print("-> OK")
print("")

print("You can start using digikam!")
print("----------------------------")
print("")
print("")
print("Statistics:")
print("   Number of tags: %d" % stats_number_of_tags)
print("   Number of pictures sucessfully imported: %d" % stats_number_of_picures)
print("   Number of pictures not imported: %d" % stats_number_of_picures_error)


# Sadly, probably nobody would "get" these...
#print "This was a triumph"
#print "I'm making a note here"
#print "HUGE SUCCESS"
#print "It's hard to overstate"
#print "my satisfaction"





fspot_session.flush()
fspot_session.rollback()

digikam_session.flush()
if dry_run:
    digikam_session.rollback()
    print("Doing a ROLLBACK! Digikam database will be reverted to original state")
else:
    digikam_session.commit()
